/*
 * Fifo-attached Serial Interface (FSI) support for SH7724
 *
 * Copyright (C) 2009 Renesas Solutions Corp.
 * Kuninori Morimoto <morimoto.kuninori@renesas.com>
 *
 * Based on ssi.c
 * Copyright (c) 2007 Manuel Lauss <mano@roarinelk.homelinux.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/list.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/initval.h>
#include <sound/soc.h>
#include <sound/pcm_params.h>
#include <asm/io.h>

#include <asm/dma.h>
#include <asm/dma-sh.h>
#include "../../../../drivers/dma/shdma.h"

#define DO_FMT		0x0000
#define DOFF_CTL	0x0004
#define DOFF_ST		0x0008
#define DI_FMT		0x000C
#define DIFF_CTL	0x0010
#define DIFF_ST		0x0014
#define CKG1		0x0018
#define CKG2		0x001C
#define DIDT		0x0020
#define DODT		0x0024
#define MUTE_ST		0x0028

#define INT_ST		0x0200
#define IEMSK		0x0204
#define IMSK		0x0208
#define MUTE		0x020C
#define CLK_RST		0x0210
#define SOFT_RST	0x0214

/* DO_FMT */
/* DI_FMT */
#define CR_FMT(param) ((param) << 4)
# define CR_MONO	0x0
# define CR_MONO_D	0x1
# define CR_PCM		0x2
# define CR_I2S		0x3
# define CR_TDM		0x4
# define CR_TDM_D	0x5

/* DOFF_CTL */
/* DIFF_CTL */
#define FIFO_CLR	0x00000001

/* DOFF_ST */
#define ERR_OVER	0x00000010
#define ERR_UNDER	0x00000001

/* CKG1 */
#define DIMD		0x00000010	/* DIIS module */
#define DOMD		0x00000001

/* CKG2 */
#define LRM		0x00001000
#define BRM		0x00000100
#define LRS		0x00000010
#define BRS		0x00000001

/* MUTE_ST */
/* 1 - 8 */
#define MUTE_OUT(x) ((1 << x) << 8)
#define MUTE_IN(x)  ((1 << x) << 0)

/* INT_ST */
/* IEMSK */
/* IMSK */
#define IN_B		0x00001000
#define OUT_B		0x00000100
#define IN_A		0x00000010
#define OUT_A		0x00000001

/* MUTE */
#define MUTE_B_IN_MASK	0x00003000
#define MUTE_B_IN	0x00002000
#define MUTE_B_IN_Z	0x00001000

#define MUTE_B_OUT_MASK	0x00000300
#define MUTE_B_OUT	0x00000200
#define MUTE_B_OUT_Z	0x00000100

#define MUTE_A_IN_MASK	0x00000030
#define MUTE_A_IN	0x00000020
#define MUTE_A_IN_Z	0x00000010

#define MUTE_A_OUT_MASK	0x00000003
#define MUTE_A_OUT	0x00000002
#define MUTE_A_OUT_Z	0x00000001

/* CLK_RST */
#define B_CLK		0x00000010
#define A_CLK		0x00000001

/* SOFT_RST */
#define B_SOFT_RESET_CANCEL	0x00001000
#define A_SOFT_RESET_CANCEL	0x00000100
#define IRQ_SOFT_RESET_CANCEL	0x00000010
#define ALL_SOFT_RESET_CANCEL	0x00000001


#include "fsi.h"
struct fsi_priv fsi_cpu_data[] = {
#if defined(CONFIG_CPU_SUBTYPE_SH7724)
	{
		.mmio	= 0xFE3C0000,
	},
/* this time A only
	{
		.mmio	= 0xFE3C0040,
	},
*/
#else
#error "Unsupported SuperH SoC"
#endif
};

static int fsi_reg_write(struct fsi_priv *fsi, u32 reg, u32 data)
{
	/* fsi register is 24bit */
	data &= 0x00ffffff;

	return ctrl_outl(data, (fsi->mmio + reg));
}

static u32 fsi_reg_read(struct fsi_priv *fsi, u32 reg)
{
	return ctrl_inl(fsi->mmio + reg);
}

static u32 fsi_reg_mask_set(struct fsi_priv *fsi, u32 reg, u32 mask, u32 data)
{
	u32 val = fsi_reg_read(fsi, reg);

	val &= ~mask;
	val |= data & mask;

	return fsi_reg_write(fsi, reg, val);
}

static void fsi_soft_reset(struct fsi_priv *fsi)
{
	u32 status = fsi_reg_read(fsi, SOFT_RST);

	/* port AB reset */
	status &= 0x000000ff;
	fsi_reg_write(fsi, SOFT_RST, status);
	mdelay(10);

	/* soft reset */
	status &= 0x000000f0;
	fsi_reg_write(fsi, SOFT_RST, status);
	status |= 0x00000001;
	fsi_reg_write(fsi, SOFT_RST, status);
	mdelay(10);
}

/* for debug */
static void fsi_print_status(struct fsi_priv *fsi)
{
	u32 val = fsi_reg_read(fsi, DOFF_ST);

	switch( 0x3 & (val >> 20)) {
	case 0: printk("Empty"); break;
	case 1: printk("low 1/2"); break;
	case 2: printk("over 1/2"); break;
	default:printk("full");
	}

	printk("- sample size = %d - ", (val >> 8) & 0x1ff);

	if (val & ERR_OVER)
		printk("over flow error ");

	if (val & ERR_UNDER)
		printk("under flow error");

	printk("\n");
}

static void fsi_stream_push(struct fsi_priv *fsi,
			    struct snd_pcm_substream *substream)
{
	struct fsi_stream_list *list;

	list = kzalloc(sizeof(*list), GFP_KERNEL);
	if (!list) {
		printk("kzalloc error %d\n", __LINE__);
		return;
	}

	INIT_LIST_HEAD(&list->next);
	list->substream = substream;
	list_add_tail(&list->next, &fsi->queue);

	list->alsa_byte_offset	= 0;
	list->alsa_buffer_size	= substream->runtime->buffer_size;
	list->alsa_total	= list->alsa_buffer_size * substream->runtime->periods;
	list->alsa_periods	= 0;
}

static struct fsi_stream_list* fsi_stream_pop(struct fsi_priv *fsi)
{
	struct fsi_stream_list *stream;

	if (list_empty(&fsi->queue))
		return NULL;

	stream = list_entry(fsi->queue.next, typeof(*stream), next);
	if (stream) {
		struct snd_pcm_runtime *runtime = stream->substream->runtime;
		struct sh7724_pcm *pcm = runtime->private_data;

		list_del(&stream->next);
		kfree(stream);
		kfree(pcm->fsi_raw_data);
	}

	return stream;
}

static int fsi_data_push(struct fsi_priv *fsi)
{
	struct snd_pcm_runtime *runtime;
	struct fsi_stream_list *stream;
	struct sh7724_pcm *pcm;
	u32 status;
	u32* fsi_dma_start;
	int fsi_word_send;
	int fsi_word_free;
	int alsa_word_index;
	int alsa_word_send;

	if (!fsi) {
		printk("--fsi NULL\n");
		return -1;
	}

get_stream:
	if (list_empty(&fsi->queue))
		return 0;

	stream = list_entry(fsi->queue.next, typeof(*stream), next);
	if (!stream) {
		printk("--substrem NULL\n");
		return -1;
	}

	/* FSI FIFO has limit.
	 * So, this driver can not send periods data at a time
	 */
	if (stream->alsa_byte_offset >=
	    stream->alsa_buffer_size * (stream->alsa_periods + 1)) {
		/* complete */
		snd_pcm_period_elapsed(stream->substream);
		stream->alsa_periods++;
	}

	if (stream->alsa_byte_offset >= stream->alsa_total) {
		/* finish operation */
		stream = fsi_stream_pop(fsi);
		goto get_stream;
	}

	if (!stream->substream) {
		printk("--substrem NULL\n");
		return -1;
	}

	if (!stream->substream->runtime) {
		printk("--runtime NULL\n");
		return -1;
	}

	runtime = stream->substream->runtime;
	pcm = runtime->private_data;

	/*
	 * following operation is tricky
	 * see comment of fsi_trigger
	 * */

	/*
	 * FSI use FIFO which have 256 column.
	 * 1 column is for 24bit data, 32bit access
	 * */

	/*
	  fsi_dma_start :	dma_start position for FSI
				it is 32bit alignment data
	  fsi_word_free:	FIFO free area size for FSI
	  alsa_word_index:	word index for alsa
				it is 16bit alignment data index
	  alsa_word_send:	send data size for alsa
				for alsa : 16bit data
				for FSI  : 32bit data
	*/

	/* get dma start position for FSI */
	alsa_word_index = stream->alsa_byte_offset / 2;
	fsi_dma_start = pcm->fsi_raw_data + alsa_word_index;

	/* get send size for alsa */
	alsa_word_send = (stream->alsa_total - stream->alsa_byte_offset) / 2;
	if (alsa_word_send > 256)
		alsa_word_send = 256;

	/*  get FIFO free size */
	status = fsi_reg_read(fsi, DOFF_ST);
	fsi_word_free = 0x1ff & (status >> 8);
	fsi_word_free = 128 - fsi_word_free;
	fsi_word_free *= 2;

	/* convert alsa count -> FSI count
	 *
	 * alsa_word_send is for 16bit data (0xaabb)
	 * fsi_word_send is for 32bit data  (0x00aabb00)
	 * same size.
	 * */
	fsi_word_send = alsa_word_send;

	/* size check */
	if (fsi_word_free < fsi_word_send)
		fsi_word_send = fsi_word_free;

	/* convert alsa <--> fsi count */
	alsa_word_send = fsi_word_send;

	/* soft dma for debug*/
	/*
	for (i=0 ; i<alsa_word_send; i++)
		fsi_reg_write(fsi, DODT, *(fsi_dma_start + i));
	*/

	/* hard dma */
	dma_configure_channel(0, (SM_INC|DM_STC|0x400|TS_32|TM_BUR));
	dma_wait_for_completion(0);
	dma_write(0, (u32)fsi_dma_start, fsi->mmio + DODT, fsi_word_send * 4);

	fsi->transmitted_len = (stream->alsa_byte_offset -1) % stream->alsa_buffer_size;
	stream->alsa_byte_offset += alsa_word_send * 2;

	fsi_reg_write(fsi, DOFF_ST, status & 0x00ffff00);
	fsi_reg_write(fsi, INT_ST, 0);
	fsi_reg_mask_set(fsi, IMSK,  0x00000001, 0x00000001);
	fsi_reg_mask_set(fsi, IEMSK, 0x00000001, 0x00000001);

	return 0;
}

static irqreturn_t fsi_interrupt(int irq, void *data)
{
	struct fsi_priv *fsi = data;
	u32 status;

	fsi_reg_mask_set(fsi, IMSK,   0x00000001, 0x00000000);
	fsi_reg_mask_set(fsi, IEMSK,  0x00000001, 0x00000000);

	fsi_data_push(fsi);

	status = fsi_reg_read(fsi, SOFT_RST) & ~0x00000010;
	fsi_reg_write(fsi, SOFT_RST, status);
	fsi_reg_write(fsi, SOFT_RST, status | 0x00000010);

	fsi_reg_write(fsi, INT_ST, 0x0000000);

	return IRQ_HANDLED;
}

/*
 * track usage of the FSI; it is simplex-only so prevent attempts of
 * concurrent playback + capture. FIXME: any locking required?
 */
static int fsi_startup(struct snd_pcm_substream *substream,
		       struct snd_soc_dai *dai)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct fsi_priv *fsi = &fsi_cpu_data[rtd->dai->cpu_dai->id];
	int err = 0;
	u32 ctrl;

	INIT_LIST_HEAD(&fsi->queue);
	spin_lock_init(&fsi->lock);

	/* hard ware reset clear*/
	fsi_soft_reset(fsi);
	fsi_print_status(fsi);

	/* ckg1 ckg2 */
	fsi_reg_write(fsi, CKG1, 0);
	//fsi_reg_write(fsi, CKG2, 0);
	fsi_reg_write(fsi, CKG2, 0x1);

	/* do fmt, di fmt */
	fsi_reg_write(fsi, DO_FMT, CR_FMT(CR_PCM));
	fsi_reg_write(fsi, DI_FMT, CR_FMT(CR_PCM));

	/* clear clk reset */
	fsi_reg_mask_set(fsi, CLK_RST, A_CLK, A_CLK); /* A port only now */

	/* irq mask setup */
	fsi_reg_mask_set(fsi, IMSK,  0x00000001, 0x00000000);
	ctrl = 1<<20;
	fsi_reg_write(fsi, DOFF_CTL, ctrl);

	fsi_reg_mask_set(fsi, IEMSK, 0x00000001, 0x00000001);
	fsi_reg_write(fsi, DOFF_CTL, ctrl | 1);

	/* irq setting */
	err = request_irq(108, &fsi_interrupt, IRQF_DISABLED, "fsi", fsi);
	if (err) {
		printk("irq error\n");
		return err;
	}

	return 0;
}

static void fsi_shutdown(struct snd_pcm_substream *substream,
			 struct snd_soc_dai *dai)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct fsi_priv *fsi = &fsi_cpu_data[rtd->dai->cpu_dai->id];

	free_irq(108, fsi);
	fsi_reg_mask_set(fsi, CLK_RST, A_CLK, 0);

	fsi_print_status(fsi);

	while (fsi_stream_pop(fsi));

}

static int fsi_trigger(struct snd_pcm_substream *substream, int cmd,
		       struct snd_soc_dai *dai)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct fsi_priv *fsi = &fsi_cpu_data[rtd->dai->cpu_dai->id];
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct sh7724_pcm *pcm = runtime->private_data;
	u16* dma_start = (u16*)runtime->dma_area;
	u32 snd;
	int i,len;

	switch (cmd) {

	/* this time START only support */
	case SNDRV_PCM_TRIGGER_START:
		fsi_stream_push(fsi, substream);

		/* 16bit -> 32bit
		 *
		 * FSI have FIFO which size is 32bit.
		 * but valid data size is 24bit.
		 * this time I use 16bit data.
		 * this time codec get 24bit data,
		 * but begining 16bit data is valid data.
		 *
		 * wav          data: 0xaabb     (16bit)
		 * FSI   input  data: 0x00ccddee (32bit)
		 *       output data: 0xccddee   (24bit)
		 * codec valid  data: 0xffgg00   (24bit)
		 *
		 * I use dma.
		 * So, plastic operation data is needed for FSI
		 *
		 * following operation mean
		 *
		 * dma_area:
		 * aabb ccdd eeff gghh ...
		 *
		 * pcm->fsi_raw_data:
		 * 00aabb00 00ccdd00 00eeff00 00gghh00...
		 *
		 * codec get:
		 * aabb00 ccdd00 eehh00...
		 *
		 * */

		len = (runtime->buffer_size * runtime->periods);
		/* should I use
		 * frames_to_bytes(runtime, runtime->period_size) ?
		 * */

		/* current len is for 16bit data length
		 * fsi_raw_data needs 32bit data length
		 * */
		pcm->fsi_raw_data = kmalloc(len*2, GFP_ATOMIC);
		if (!pcm->fsi_raw_data)
			return -ENOMEM;

		/* sometimes this driver get BUG_ON
		 * when above kmalloc is called.
		 * but why ?
		 * I use GFP_ATOMIC option, because
		 * http://www.kernel.org/pub/linux/kernel/people/tiwai/docs/writing-an-alsa-driver/ch05s06.html#pcm-interface-operators-trigger-callback
		 * said it can not sleep.
		 * BUG_ON occur when I use GFP_KERNEL.
		 * */

		/* data convert */
		for (i=0 ; i<len/2; i++) {
			snd = (u32)dma_start[i];
			pcm->fsi_raw_data[i] = snd << 8;
		}

		fsi_data_push(fsi);
	}

	return 0;
}

#define FSI_RATES SNDRV_PCM_RATE_8000_96000

#define FSI_FMTS	\
	(SNDRV_PCM_FMTBIT_S8      | SNDRV_PCM_FMTBIT_U8      |	\
	 SNDRV_PCM_FMTBIT_S16_LE  | SNDRV_PCM_FMTBIT_U16_LE  |	\
	 SNDRV_PCM_FMTBIT_S20_3LE | SNDRV_PCM_FMTBIT_U20_3LE |	\
	 SNDRV_PCM_FMTBIT_S24_3LE | SNDRV_PCM_FMTBIT_U24_3LE |	\
	 SNDRV_PCM_FMTBIT_S32_LE  | SNDRV_PCM_FMTBIT_U32_LE)

static struct snd_soc_dai_ops fsi_dai_ops = {
	.startup	= fsi_startup,
	.shutdown	= fsi_shutdown,
	.trigger	= fsi_trigger,
};

struct snd_soc_dai sh4_fsi_dai[] = {
	{
		.name			= "FSI0",
		.id			= 0,
		.playback = {
			.rates		= FSI_RATES,
			.formats	= FSI_FMTS,
			.channels_min	= 2,
			.channels_max	= 2,
		},
/* this time not support
		.capture = {
			.rates		= FSI_RATES,
			.formats	= FSI_FMTS,
			.channels_min	= 1,
			.channels_max	= 8,
		},
*/
		.ops = &fsi_dai_ops,
	}
};
EXPORT_SYMBOL_GPL(sh4_fsi_dai);

static int __init sh4_fsi_init(void)
{
	return snd_soc_register_dais(sh4_fsi_dai, ARRAY_SIZE(sh4_fsi_dai));
}
module_init(sh4_fsi_init);

static void __exit sh4_fsi_exit(void)
{
	snd_soc_unregister_dais(sh4_fsi_dai, ARRAY_SIZE(sh4_fsi_dai));
}
module_exit(sh4_fsi_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SuperH onchip FSI audio driver");
MODULE_AUTHOR("Kuninori Morimoto <morimoto.kuninori@renesas.com>");
